package io.data2viz.charts.event

import io.data2viz.charts.chart.Chart
import io.data2viz.charts.core.Datum
import io.data2viz.charts.core.MarkIndexedDatum
import io.data2viz.charts.core.MarkIndexedDomain
import io.data2viz.charts.core.SelectedDatum
import io.data2viz.geom.Rect
import io.data2viz.math.Percent
import kotlin.jvm.JvmName

/**
 * Interface to control a [Chart]: highlight or select elements, zoom the view or pan it.
 */
public interface ChartController<DOMAIN> {

    /**
     * Highlight the given DOMAIN objects.
     *
     * @param domains A [Collection] of DOMAIN objects to highlight on screen.
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("highlightDomains")
    public fun highlight(domains: Collection<DOMAIN>)

    /**
     * Highlight the given [Datum]s.
     *
     * @param data A [Collection] of [Datum] to highlight on screen.
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("highlightData")
    public fun highlight(data: Collection<Datum<DOMAIN>>)

    /**
     * Highlight the given DOMAIN objects for the given [Mark]s indexes.
     *
     * @param domains A [Collection] of [MarkIndexedDomain] to highlight on screen. Each contain the information about
     * the datum and the mark to select, only use this function when you have multiple marks.
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("highlightForDomains")
    public fun highlightFor(domainsPerMark: Map<Int,Collection<DOMAIN>>)

    /**
     * Select the given [Datum]s for the given [Mark]s indexes.
     *
     * @param data A [Collection] of [MarkIndexedDatum] to highlight on screen. Each contain the information about
     * the datum and the mark to select, only use this function when you have multiple marks.
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName(name = "highlightForData")
    public fun highlightFor(selectedData: Collection<SelectedDatum<DOMAIN>>)


    /**
     * Select the given DOMAIN objects.
     *
     * @param domains A [Collection] of DOMAIN objects to select on screen.
     * @param selectionMarkIndexes A list of mark indexes used to restrict selection only for these marks
     *  (default to emptyList).
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("selectDomains")
    public fun select(domains: Collection<DOMAIN>)

    /**
     * Select the given [Datum]s.
     *
     * @param data A [Collection] of [Datum] to highlight on screen.
     * @param selectionMarkIndexes A list of mark indexes used to restrict selection only for these marks
     *  (default to emptyList).
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("selectData")
    public fun select(data: Collection<Datum<DOMAIN>>)

    /**
     * Add the given DOMAIN objects to the current selection.
     *
     * @param domains A [Collection] of DOMAIN objects to select on screen.
     * @param selectionMarkIndexes A list of mark indexes used to restrict selection only for these marks
     *  (default to emptyList).
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("addToSelectionDomains")
    public fun addToSelection(domains: Collection<DOMAIN>)

    /**
     * Add the given [Datum]s to the current selection.
     *
     * @param data A [Collection] of [Datum] to add to selection on screen.
     * @param selectionMarkIndexes A list of mark indexes used to restrict selection only for these marks
     *  (default to emptyList).
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("addToSelectionData")
    public fun addToSelection(data: Collection<SelectedDatum<DOMAIN>>)

    /**
     * Remove the given DOMAIN objects from the current selection.
     *
     * @param domains A [Collection] of DOMAIN objects to unselect on screen.
     * @param selectionMarkIndexes A list of mark indexes used to restrict selection only for these marks
     *  (default to emptyList).
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("removeFromSelectionDomains")
    public fun removeFromSelection(domains: Collection<DOMAIN>)

    /**
     * Remove the given [Datum]s from the current selection.
     *
     * @param data A [Collection] of [Datum] to unselect on screen.
     * @param selectionMarkIndexes A list of mark indexes used to restrict selection only for these marks
     *  (default to emptyList).
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("removeFromSelectionData")
    public fun removeFromSelection(data: Collection<SelectedDatum<DOMAIN>>)

    /**
     * Select the given DOMAIN objects for the given [Mark]s indexes.
     *
     * @param domains A [Map] of [DOMAIN] to select on screen.
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName("selectForDomains")
    public fun selectFor(domainsPerMark: Map<Int,Collection<DOMAIN>>)

    /**
     * Select the given [Datum]s for the given [Mark]s indexes.
     *
     * @param data A [Map] of [Datum] to select on screen.
     */
    @Suppress("INAPPLICABLE_JVM_NAME")
    @JvmName(name = "selectForData")
    public fun selectFor(selectedData: Collection<SelectedDatum<DOMAIN>>)

    /**
     * Display the [Cursor] on a given DOMAIN object and a given Mark. Set to null to hide cursor.
     *
     * @param domain A [DOMAIN] object to place the cursor on.
     * @param markIndex The 0-based index of the mark for which you want to display the cursor (default 0).
     */
    public fun setCursorFor(domain: DOMAIN?, markIndex: Int = 0)

    /**
     * Display the [Cursor] on a given Datum and a given Mark. Set to null to hide cursor.
     *
     * @param selectedDatum A [SelectedDatum] to place the cursor on.
     */
    public fun setCursorFor(selectedDatum: SelectedDatum<DOMAIN>?)

    /**
     * Zoom the "viewport" of the [Chart] by a certain percentage in X or Y direction, knowing the zoom origin.
     * For example zoom(50%, 50%, 110%, 110%) will zoom by a ratio 1.1 in X and Y directions around the middle of the
     * charting zone.
     *
     * @param zoomAction The full parameters of the zoom.
     */
    public fun zoom(zoomAction: ZoomAction)

    @Deprecated("Use zoom(zoomAction)", replaceWith = ReplaceWith("zoom(ZoomAction(RelativePoint(zoomOriginX, zoomOriginY), zoomFactorX, zoomFactorY))", "io.data2viz.charts.event.ZoomAction", "io.data2viz.charts.chart.mark.RelativePoint"))
    public fun zoom(zoomOriginX: Percent, zoomOriginY: Percent, zoomFactorX: Percent, zoomFactorY: Percent)

    // TODO introduce PanAction data class
    /**
     * Move the "viewport" of the [Chart] by a certain percentage in X or Y direction.
     * For example pan(-10%, 20%) will move the charting zone 10% to the left and 20% up.
     *
     * @param panX The percentage of panning in [Percent] along the X-axis.
     * @param panY The percentage of panning in [Percent] along the Y-axis.
     */
    public fun pan(panX: Percent, panY: Percent)

    /**
     * Reset the view, setting the pan and zoom to their origin values.
     */
    public fun viewReset()

    /**
     * Clear any selection box on screen.
     */
    public fun clearSelection()

    /**
     * Draw a selection box on screen.
     */
    public fun displaySelection(selectedZone: Rect?)

    /**
     * Draw an "add to selection" box on screen.
     */
    public fun displaySelectionAdd(selectedZone: Rect?)

    /**
     * Draw a "remove from selection" box on screen.
     */
    public fun displaySelectionRemove(selectedZone: Rect?)

    /**
     * Draw a "zoom selection" box on screen.
     */
    public fun displaySelectionZoom(selectedZone: Rect?)

    /**
     * The default event handling lambda of your chart, it is the default value of your [getChartActionOnUserEvent].
     *
     * Useful when you want to override some specific event and bypass the others, you can change the
     * [getChartActionOnUserEvent] value for your event, and bypass all the others by calling
     * [defaultChartActionOnUserEvent] with your [EventContext].
     */
    public val defaultChartActionOnUserEvent: EventContext<DOMAIN>.() -> Unit

    /**
     * Change this to handle specific events and override some standard behaviors.
     * Use the default value [defaultChartActionOnUserEvent] to bypass and use the default behavior.
     *
     * For example, if you want to zoom-in (by 20% in both X and Y) when CTRL-clicking on some point:
     * ```
     * getChartActionOnUserEvent = {
     *      if (event.type == EventType.Click && event.ctrlKey) {
     *          chart.zoom(ZoomAction(currentPosition, 120.pct, 120.pct))
     *      } else {
     *          defaultChartActionOnUserEvent(this)
     *      }
     * }
     * ```
     */
    public var getChartActionOnUserEvent: EventContext<DOMAIN>.() -> Unit
}
